# Enthought library imports.
from envisage.ui.tasks.api import PreferencesPane, TaskFactory
from apptools.preferences.api import PreferencesHelper
from traits.api import Bool, Dict, Enum, List, Str
from traitsui.api import EnumEditor, HGroup, VGroup, Item, Label, View


class AttractorsPreferences(PreferencesHelper):
    """ The preferences helper for the Attractors application.
    """

    #### 'PreferencesHelper' interface ########################################

    # The path to the preference node that contains the preferences.
    preferences_path = "example.attractors"

    #### Preferences ##########################################################

    # The task to activate on app startup if not restoring an old layout.
    default_task = Str

    # Whether to always apply the default application-level layout.
    # See TasksApplication for more information.
    always_use_default_layout = Bool


class AttractorsPreferencesPane(PreferencesPane):
    """ The preferences pane for the Attractors application.
    """

    #### 'PreferencesPane' interface ##########################################

    # The factory to use for creating the preferences model object.
    model_factory = AttractorsPreferences

    #### 'AttractorsPreferencesPane' interface ################################

    task_map = Dict(Str, Str)

    view = View(
        VGroup(
            HGroup(
                Item("always_use_default_layout"),
                Label("Always use the default active task on startup"),
                show_labels=False,
            ),
            HGroup(
                Label("Default active task:"),
                Item(
                    "default_task", editor=EnumEditor(name="handler.task_map")
                ),
                enabled_when="always_use_default_layout",
                show_labels=False,
            ),
            label="Application startup",
        ),
        resizable=True,
    )

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _task_map_default(self):
        return dict(
            (factory.id, factory.name)
            for factory in self.dialog.application.task_factories
        )



""" A simple example of a GUIApplication which wraps a TraitsUI """

from traits.api import HasTraits, Str, Int, Enum, Instance, on_trait_change
from envisage.api import Plugin
from envisage.ui.gui_application import GUIApplication
from traitsui.api import View, Item, OKCancelButtons


class Person(HasTraits):
    """ A typical traits model object """

    name = Str("John Doe")

    age = Int(21)

    gender = Enum("Male", "Female")

    view = View(
        Item("name"), Item("age"), Item("gender"), buttons=OKCancelButtons,
    )


class PersonViewPlugin(Plugin):
    """ The 'Person View' plugin.

    This plugin waits for the application to start, and then creates a traits
    UI.

    """

    ui = Instance("traitsui.ui.UI")

    @on_trait_change("application:application_initialized")
    def on_application_start(self):
        """ Start the UI. """

        person = Person()

        # keep a reference to the ui object to avoid garbage collection
        self.ui = person.edit_traits()


# Application entry point.
if __name__ == "__main__":

    # Create the application.
    application = GUIApplication(
        id="person_view", plugins=[PersonViewPlugin()]
    )

    # Run it!
    application.run()



# (C) Copyright 2007-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" Tests for the resource manager. """


# Standard library imports.
import unittest
from urllib.error import HTTPError
import urllib.request

from io import StringIO

# Major package imports.
from pkg_resources import resource_filename

# Enthought library imports.
from envisage.resource.api import ResourceManager
from envisage.resource.api import NoSuchResourceError

# Module to patch urlopen in during testing.
url_library = urllib.request


# This module's package.
PKG = "envisage.resource.tests"


# mimics `urlopen` for some tests.
# In setUp it replaces `urlopen` for some tests,
# and in tearDown, the regular `urlopen` is put back into place.
def stubout_urlopen(url):
    if "bogus" in url:
        raise HTTPError(url, "404", "No such resource", "", None)

    elif "localhost" in url:
        return StringIO("This is a test file.\n")

    else:
        raise ValueError("Unexpected URL %r in stubout_urlopen" % url)


class ResourceManagerTestCase(unittest.TestCase):
    """ Tests for the resource manager. """

    ###########################################################################
    # 'TestCase' interface.
    ###########################################################################

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """

        self.stored_urlopen = url_library.urlopen
        url_library.urlopen = stubout_urlopen

    def tearDown(self):
        """ Called immediately after each test method has been called. """

        url_library.urlopen = self.stored_urlopen

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_file_resource(self):
        """ file resource """

        rm = ResourceManager()

        # Get the filename of the 'api.py' file.
        filename = resource_filename("envisage.resource", "api.py")

        # Open a file resource.
        f = rm.file("file://" + filename)
        self.assertNotEqual(f, None)
        contents = f.read()
        f.close()

        # Open the api file via the file system.
        with open(filename, "rb") as g:
            self.assertEqual(g.read(), contents)

    def test_no_such_file_resource(self):
        """ no such file resource """

        rm = ResourceManager()

        # Open a file resource.
        with self.assertRaises(NoSuchResourceError):
            rm.file("file://../bogus.py")

    def test_package_resource(self):
        """ package resource """

        rm = ResourceManager()

        # Open a package resource.
        f = rm.file("pkgfile://envisage.resource/api.py")
        self.assertNotEqual(f, None)
        contents = f.read()
        f.close()

        # Get the filename of the 'api.py' file.
        filename = resource_filename("envisage.resource", "api.py")

        # Open the api file via the file system.
        with open(filename, "rb") as g:
            self.assertEqual(g.read(), contents)

    def test_no_such_package_resource(self):
        """ no such package resource """

        rm = ResourceManager()

        # Open a package resource.
        with self.assertRaises(NoSuchResourceError):
            rm.file("pkgfile://envisage.resource/bogus.py")

        with self.assertRaises(NoSuchResourceError):
            rm.file("pkgfile://completely.bogus/bogus.py")

    def test_http_resource(self):
        """ http resource """

        # Open an HTTP document resource.
        rm = ResourceManager()

        f = rm.file("http://localhost:1234/file.dat")
        self.assertNotEqual(f, None)
        contents = f.read()
        f.close()

        self.assertEqual(contents, "This is a test file.\n")

    def test_no_such_http_resource(self):
        """ no such http resource """

        # Open an HTTP document resource.
        rm = ResourceManager()

        with self.assertRaises(NoSuchResourceError):
            rm.file("http://localhost:1234/bogus.dat")

    def test_unknown_protocol(self):
        """ unknown protocol """

        # Open an HTTP document resource.
        rm = ResourceManager()

        with self.assertRaises(ValueError):
            rm.file("bogus://foo/bar/baz")



# (C) Copyright 2007-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
"""
- :class:`~.StartQtConsoleAction`
- :class:`~.InternalIPKernel`
- :class:`~.IPythonKernelPlugin`
- :attr:`~.IPYTHON_KERNEL_PROTOCOL`
- :class:`~.IPythonKernelUIPlugin`
"""
from envisage.plugins.ipython_kernel.actions import StartQtConsoleAction
from envisage.plugins.ipython_kernel.internal_ipkernel import InternalIPKernel
from envisage.plugins.ipython_kernel.ipython_kernel_plugin import (
    IPythonKernelPlugin, IPYTHON_KERNEL_PROTOCOL,
)
from envisage.plugins.ipython_kernel.ipython_kernel_ui_plugin import (
    IPythonKernelUIPlugin,
)



# (C) Copyright 2007-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
from .i_action_set import IActionSet
from .i_action_manager_builder import IActionManagerBuilder

from .abstract_action_manager_builder import AbstractActionManagerBuilder
from .action import Action
from .action_set import ActionSet
from .group import Group
from .menu import Menu
from .tool_bar import ToolBar



# (C) Copyright 2007-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" An implementation of weak references that works for bound methods.

This code is based on the code in the Python Cookbook, but you can call `ref`
for objects that are *not* bound methods too, in which case it just returns a
standard `weakref.ref`.

Weak references to bound methods are cached so that `ref(x) is ref(x)` as for
standard weakrefs, and the `ref` class defined here is therefore intended to be
used as a drop-in replacement for 'weakref.ref'.

"""

# Standard library imports.
import warnings
import weakref


class ref(object):
    """An implementation of weak references that works for bound methods and
    caches them.

    If ``object`` is a bound method, returns a ``weakref.WeakMethod`` for that
    method. This ensures that the method is kept alive for the lifetime of the
    object that it's bound to.

    For any other ``object``, a normal ``weakref.ref`` is returned.

    .. deprecated:: 5.0.0

    """
    _cache = weakref.WeakKeyDictionary()

    def __new__(cls, obj, callback=None):
        warnings.warn(
            message=(
                "safeweakref.ref is deprecated, and will be removed in a "
                "future version of Envisage"
            ),
            category=DeprecationWarning,
            stacklevel=2,
        )

        if getattr(obj, "__self__", None) is not None:  # Bound method
            # Caching
            func_cache = cls._cache.setdefault(obj.__self__, {})
            self = func_cache.get(obj.__func__)
            if self is None:
                self = weakref.WeakMethod(obj, callback)
                func_cache[obj.__func__] = self
            return self
        else:
            return weakref.ref(obj, callback)



""" The default implementation of the 'IMessage' interface. """


# Enthought library imports.
from traits.api import HasTraits, Str, provides

# Local imports.
from acme.motd.i_message import IMessage


@provides(IMessage)
class Message(HasTraits):
    """ The default implementation of the 'IMessage' interface. """

    # The author of the message.
    author = Str

    # The text of the message.
    text = Str



""" The 'Message of the Day' interface. """


# Enthought library imports.
from traits.api import Interface


class IMOTD(Interface):
    """ The 'Message of the Day' interface. """

    def motd(self):
        """ Return the message of the day.

        Returns an object that implements the 'IMessage' interface.

        """



# (C) Copyright 2007-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" A trait type used to declare and access extension points. """


# Standard library imports.
from functools import wraps
import inspect
import warnings
import weakref

# Enthought library imports.
from traits.api import List, TraitType, Undefined, provides
from traits.trait_list_object import TraitList

# Local imports.
from .i_extension_point import IExtensionPoint


# Exception message template.
INVALID_TRAIT_TYPE = (
    'extension points must be "List"s e.g. List, List(Int)'
    " but a value of %s was specified."
)


# Even though trait types do not themselves have traits, we can still
# declare that we implement an interface.
@provides(IExtensionPoint)
class ExtensionPoint(TraitType):
    """ A trait type used to declare and access extension points.

    Note that this is a trait *type* and hence does *NOT* have traits itself
    (i.e. it does *not* inherit from 'HasTraits').

    """

    ###########################################################################
    # 'ExtensionPoint' *CLASS* interface.
    ###########################################################################

    @staticmethod
    def bind(obj, trait_name, extension_point_id):
        """ Create a binding to an extension point. """

        from .extension_point_binding import bind_extension_point

        return bind_extension_point(obj, trait_name, extension_point_id)

    @staticmethod
    def connect_extension_point_traits(obj):
        """ Connect all of the 'ExtensionPoint' traits on an object. """

        for trait_name, trait in obj.traits(__extension_point__=True).items():
            trait.trait_type.connect(obj, trait_name)

    @staticmethod
    def disconnect_extension_point_traits(obj):
        """ Disconnect all of the 'ExtensionPoint' traits on an object. """

        for trait_name, trait in obj.traits(__extension_point__=True).items():
            trait.trait_type.disconnect(obj, trait_name)

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, trait_type=List, id=None, **metadata):
        """ Constructor. """

        # We add '__extension_point__' to the metadata to make the extension
        # point traits easier to find with the 'traits' and 'trait_names'
        # methods on 'HasTraits'.
        metadata["__extension_point__"] = True
        super().__init__(**metadata)

        # The trait type that describes the extension point.
        #
        # If we are handed a trait type *class* e.g. List, instead of a trait
        # type *instance* e.g. List() or List(Int) etc, then we instantiate it.
        if inspect.isclass(trait_type):
            trait_type = trait_type()

        # Currently, we only support list extension points (we may in the
        # future want to allow other collections e.g. dictionaries etc).
        if not isinstance(trait_type, List):
            raise TypeError(INVALID_TRAIT_TYPE % trait_type)

        self.trait_type = trait_type

        # The Id of the extension point.
        if id is None:
            raise ValueError("an extension point must have an Id")

        self.id = id

        # A dictionary that is used solely to keep a reference to all extension
        # point listeners alive until their associated objects are garbage
        # collected.
        #
        # Dict(weakref.ref(Any), Dict(Str, Callable))
        self._obj_to_listeners_map = weakref.WeakKeyDictionary()

    def __repr__(self):
        """ String representation of an ExtensionPoint object """
        return "ExtensionPoint(id={!r})".format(self.id)

    ###########################################################################
    # 'TraitType' interface.
    ###########################################################################

    def get(self, obj, trait_name):
        """ Trait type getter. """
        cache_name = _get_cache_name(trait_name)
        if cache_name not in obj.__dict__:
            _update_cache(obj, trait_name)

        value = obj.__dict__[cache_name]
        # validate again
        self.trait_type.validate(obj, trait_name, value[:])
        return value

    def set(self, obj, name, value):
        """ Trait type setter. """

        extension_registry = self._get_extension_registry(obj)

        # Note that some extension registry implementations may not support the
        # setting of extension points (the default, plugin extension registry
        # for exxample ;^).
        extension_registry.set_extensions(self.id, value)

    ###########################################################################
    # 'ExtensionPoint' interface.
    ###########################################################################

    def connect(self, obj, trait_name):
        """ Connect the extension point to a trait on an object.

        This allows the object to react when contributions are added or
        removed from the extension point.

        fixme: It would be nice to be able to make the connection automatically
        but we would need a slight tweak to traits to allow the trait type to
        be notified when a new instance that uses the trait type is created.

        """

        def listener(extension_registry, event):
            """ Listener called when an extension point is changed.

            Parameters
            ----------
            extension_registry : IExtensionRegistry
                Registry that maintains the extensions.
            event : ExtensionPointChangedEvent
                Event created for the change.
                If the event.index is None, this means the entire extensions
                list is set to a new value. If the event.index is not None,
                some portion of the list has been modified.
            """
            if event.index is not None:
                # We know where in the list is changed.

                # Mutate the _ExtensionPointValue to fire ListChangeEvent
                # expected from observing item change.
                getattr(obj, trait_name)._sync_values(event)

                # For on_trait_change('name_items')
                obj.trait_property_changed(
                    trait_name + "_items", Undefined, event
                )

            else:
                # The entire list has changed. We reset the cache and fire a
                # normal trait changed event.
                _update_cache(obj, trait_name)

        # In case the cache was created first and the registry is then mutated
        # before this ``connect`` is called, the internal cache would be in
        # an inconsistent state. This also has the side-effect of firing
        # another change event, hence allowing future changes to be observed
        # without having to access the trait first.
        _update_cache(obj, trait_name)

        extension_registry = self._get_extension_registry(obj)

        # Add the listener to the extension registry.
        extension_registry.add_extension_point_listener(listener, self.id)

        # Save a reference to the listener so that it does not get garbage
        # collected until its associated object does.
        listeners = self._obj_to_listeners_map.setdefault(obj, {})
        listeners[trait_name] = listener

    def disconnect(self, obj, trait_name):
        """ Disconnect the extension point from a trait on an object. """

        extension_registry = self._get_extension_registry(obj)

        listener = self._obj_to_listeners_map[obj].get(trait_name)
        if listener is not None:
            # Remove the listener from the extension registry.
            extension_registry.remove_extension_point_listener(
                listener, self.id
            )

            # Clean up.
            del self._obj_to_listeners_map[obj][trait_name]

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_extension_registry(self, obj):
        """ Return the extension registry in effect for an object. """

        extension_registry = getattr(obj, "extension_registry", None)
        if extension_registry is None:
            raise ValueError(
                'The "ExtensionPoint" trait type can only be used in '
                "objects that have a reference to an extension registry "
                'via their "extension_registry" trait. '
                "Extension point Id <%s>" % self.id
            )

        return extension_registry


def _warn_if_not_internal(func):
    """ Decorator for instance methods of _ExtensionPointValue such that its
    effect is nullified if the function is not called with the _internal_use
    flag set to true.
    """

    @wraps(func)
    def decorated(object, *args, **kwargs):
        if not object._internal_use:
            warnings.warn(
                "Extension point cannot be mutated directly.",
                RuntimeWarning,
                stacklevel=2,
            )
            # This restores the existing behavior where the operation
            # is acted on a list object that is not persisted.
            return func(TraitList(iter(object)), *args, **kwargs)
        return func(object, *args, **kwargs)

    return decorated


class _ExtensionPointValue(TraitList):
    """ _ExtensionPointValue is the list being returned while retrieving the
    attribute value for an ExtensionPoint trait.

    This list returned for an ExtensionPoint acts as a proxy to query
    extensions in an ExtensionRegistry for a given extension point id. Users of
    ExtensionPoint expect to handle a list-like object, and expect to be able
    to listen to "mutation" on the list. The ExtensionRegistry remains to be
    the source of truth as to what extensions are available for a given
    extension point ID.

    Users are not expected to mutate the list directly. All mutations to
    extensions are expected to go through the extension registry to maintain
    consistency. With that, all methods for mutating the list are nullified,
    unless it is used internally.

    The requirement to support ``observe("name:items")`` means this list,
    associated with `name`, cannot be a property that gets recomputed on every
    access (enthought/traits#624), it needs to be cached. As with any
    cached quantity, it needs to be synchronized with the ExtensionRegistry.

    Note that the list can only be synchronized with the extension registry
    when the listeners are connected (see ``ExtensionPoint.connect``).

    Parameters
    ----------
    iterable : iterable
        Iterable providing the items for the list
    """

    def __new__(cls, *args, **kwargs):
        # Methods such as 'append' or 'extend' may be called during unpickling.
        # Initialize internal flag to true which gets changed back to false
        # in __init__.
        self = super().__new__(cls)
        self._internal_use = True
        return self

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Flag to control access for mutating the list. Only internal
        # code can mutate the list. See _sync_values
        self._internal_use = False

    def _sync_values(self, event):
        """ Given an ExtensionPointChangedEvent, modify the values in this list
        to match. This is an internal method only used by Envisage code.

        Parameters
        ----------
        event : ExtenstionPointChangedEvent
            Event being fired for extension point values changed (typically
            via the extension registry)
        """
        self._internal_use = True
        try:
            if isinstance(event.index, slice):
                if event.added:
                    self[event.index] = event.added
                else:
                    del self[event.index]
            else:
                slice_ = slice(
                    event.index, event.index + len(event.removed)
                )
                self[slice_] = event.added
        finally:
            self._internal_use = False

    __delitem__ = _warn_if_not_internal(TraitList.__delitem__)
    __iadd__ = _warn_if_not_internal(TraitList.__iadd__)
    __imul__ = _warn_if_not_internal(TraitList.__imul__)
    __setitem__ = _warn_if_not_internal(TraitList.__setitem__)
    append = _warn_if_not_internal(TraitList.append)
    clear = _warn_if_not_internal(TraitList.clear)
    extend = _warn_if_not_internal(TraitList.extend)
    insert = _warn_if_not_internal(TraitList.insert)
    pop = _warn_if_not_internal(TraitList.pop)
    remove = _warn_if_not_internal(TraitList.remove)
    reverse = _warn_if_not_internal(TraitList.reverse)
    sort = _warn_if_not_internal(TraitList.sort)


def _get_extensions(object, name):
    """ Return the extensions reported by the extension registry for the
    given object and the name of a trait whose type is an ExtensionPoint.

    Parameters
    ----------
    object : HasTraits
        Object on which an ExtensionPoint is defined
    name : str
        Name of the trait whose trait type is an ExtensionPoint.

    Returns
    -------
    extensions : list
        All the extensions for the extension point.
    """
    extension_point = object.trait(name).trait_type
    extension_registry = extension_point._get_extension_registry(object)

    # Get the extensions to this extension point.
    return extension_registry.get_extensions(extension_point.id)


def _get_cache_name(trait_name):
    """ Return the attribute name on the object for storing the cached
    extension point value associated with a given trait.

    Parameters
    ----------
    trait_name : str
        The name of the trait for which ExtensionPoint is defined.
    """
    return "__envisage_{}".format(trait_name)


def _update_cache(obj, trait_name):
    """ Update the internal cached value for the extension point and
    fire change event.

    Parameters
    ----------
    obj : HasTraits
        The object on which an ExtensionPoint is defined.
    trait_name : str
        The name of the trait for which ExtensionPoint is defined.
    """
    cache_name = _get_cache_name(trait_name)
    old = obj.__dict__.get(cache_name, Undefined)
    new = (
        _ExtensionPointValue(
            _get_extensions(obj, trait_name)
        )
    )
    obj.__dict__[cache_name] = new
    obj.trait_property_changed(trait_name, old, new)



""" The 'Software Quotes' plugin """


# Enthought library imports.
from envisage.api import Plugin
from traits.api import List


class SoftwareQuotesPlugin(Plugin):
    """ The 'Software Quotes' plugin. """

    #### 'IPlugin' interface ##################################################

    # The plugin's unique identifier.
    id = "acme.motd.software_quotes"

    # The plugin's name (suitable for displaying to the user).
    name = "Software Quotes"

    #### Extension points offered by this plugin ##############################

    # None

    #### Contributions to extension points made by this plugin ################

    # Messages for the 'Message Of The Day'.
    messages = List(contributes_to="acme.motd.messages")

    def _messages_default(self):
        """ Trait initializer. """

        # Only do imports when you need to!
        from acme.motd.software_quotes.messages import messages

        return messages



# (C) Copyright 2007-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" A test class used to test adapters. """


# Enthought library imports.
from traits.api import Interface


class IFoo(Interface):
    pass



# (C) Copyright 2007-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" Tests for the events fired when extension points are changed. """

# Standard library imports.
import unittest

# Local imports.
from envisage.api import ExtensionPointChangedEvent
from envisage.tests.test_application import (
    PluginA,
    PluginB,
    PluginC,
    TestApplication,
    listener,
)


class ExtensionPointChangedTestCase(unittest.TestCase):
    """ Tests for the events fired when extension points are changed. """

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """

        # Make sure that the listener contents get cleand up before each test.
        listener.obj = None
        listener.trait_name = None
        listener.old = None
        listener.new = None

    def test_set_extension_point(self):
        """ set extension point """

        a = PluginA()

        application = TestApplication(plugins=[a])
        application.start()

        # Try to set the extension point.
        with self.assertRaises(SystemError):
            setattr(a, "x", [1, 2, 3])

    def test_mutate_extension_point_no_events(self):
        """ Mutation will not emit change event for name_items """

        a = PluginA()
        b = PluginB()
        c = PluginC()

        a.on_trait_change(listener, "x_items")
        events = []
        a.observe(events.append, "x:items")

        application = TestApplication(plugins=[a, b, c])
        application.start()

        # when
        with self.assertWarns(RuntimeWarning):
            a.x.append(42)

        # then
        self.assertIsNone(listener.obj)
        self.assertEqual(len(events), 0)

    def test_mutate_extension_point_then_modify_from_registry(self):
        """ Mutating the extension point does nothing and should not cause
        subsequent change event information to become inconsistent.
        """
        a = PluginA()
        b = PluginB()
        c = PluginC()

        a.on_trait_change(listener, "x_items")
        events = []
        a.observe(events.append, "x:items")

        application = TestApplication(plugins=[a, b, c])
        application.start()

        # when
        with self.assertWarns(RuntimeWarning):
            a.x.clear()

        # then
        self.assertIsNone(listener.obj)
        self.assertEqual(len(events), 0)

        # when
        # Append a contribution.
        b.x.append(4)

        # then
        self.assertEqual(a.x, [1, 2, 3, 4, 98, 99, 100])
        self.assertEqual(len(events), 1)
        event, = events
        self.assertEqual(event.object, a.x)
        self.assertEqual(event.index, 3)
        self.assertEqual(event.added, [4])
        self.assertEqual(event.removed, [])

    def test_append(self):
        """ append """

        a = PluginA()
        a.on_trait_change(listener, "x_items")
        b = PluginB()
        c = PluginC()

        application = TestApplication(plugins=[a, b, c])
        application.start()

        # Append a contribution.
        b.x.append(4)

        # Make sure we pick up the new contribution via the application.
        extensions = application.get_extensions("a.x")
        extensions.sort()

        self.assertEqual(7, len(extensions))
        self.assertEqual([1, 2, 3, 4, 98, 99, 100], extensions)

        # Make sure we pick up the new contribution via the plugin.
        extensions = a.x[:]
        extensions.sort()

        self.assertEqual(7, len(extensions))
        self.assertEqual([1, 2, 3, 4, 98, 99, 100], extensions)

        # Make sure we got a trait event telling us that the contributions
        # to the extension point have been changed.
        self.assertEqual(a, listener.obj)
        self.assertEqual("x_items", listener.trait_name)
        self.assertEqual([4], listener.new.added)
        self.assertEqual([], listener.new.removed)
        self.assertEqual(3, listener.new.index)

    def test_append_with_observe(self):
        """ append with observe """

        a = PluginA()
        b = PluginB()
        c = PluginC()

        events = []
        a.observe(events.append, "x:items")

        application = TestApplication(plugins=[a, b, c])
        application.start()

        # Append a contribution.
        b.x.append(4)

        # then
        self.assertEqual(len(events), 1)
        event, = events
        self.assertEqual(event.object, a.x)
        self.assertEqual(event.index, 3)
        self.assertEqual(event.added, [4])
        self.assertEqual(event.removed, [])

    def test_remove(self):
        """ remove """

        a = PluginA()
        a.on_trait_change(listener, "x_items")
        b = PluginB()
        c = PluginC()

        application = TestApplication(plugins=[a, b, c])
        application.start()

        # Remove a contribution.
        b.x.remove(3)

        # Make sure we pick up the correct contributions via the application.
        extensions = application.get_extensions("a.x")
        extensions.sort()

        self.assertEqual(5, len(extensions))
        self.assertEqual([1, 2, 98, 99, 100], extensions)

        # Make sure we pick up the correct contributions via the plugin.
        extensions = a.x[:]
        extensions.sort()

        self.assertEqual(5, len(extensions))
        self.assertEqual([1, 2, 98, 99, 100], extensions)

        # Make sure we got a trait event telling us that the contributions
        # to the extension point have been changed.
        self.assertEqual(a, listener.obj)
        self.assertEqual("x_items", listener.trait_name)
        self.assertEqual([], listener.new.added)
        self.assertEqual([3], listener.new.removed)
        self.assertEqual(2, listener.new.index)

    def test_remove_with_observe(self):
        """ remove with observing items change. """

        a = PluginA()
        b = PluginB()
        c = PluginC()

        events = []
        a.observe(events.append, "x:items")

        application = TestApplication(plugins=[a, b, c])
        application.start()

        # Remove a contribution.
        b.x.remove(3)

        # then
        self.assertEqual(len(events), 1)
        event, = events
        self.assertEqual(event.object, a.x)
        self.assertEqual(event.index, 2)
        self.assertEqual(event.added, [])
        self.assertEqual(event.removed, [3])

    def test_assign_empty_list(self):
        """ assign empty list """

        a = PluginA()
        a.on_trait_change(listener, "x_items")
        b = PluginB()
        c = PluginC()

        application = TestApplication(plugins=[a, b, c])
        application.start()

        # Assign an empty list to one of the plugin's contributions.
        b.x = []

        # Make sure we pick up the correct contribution via the application.
        extensions = application.get_extensions("a.x")
        extensions.sort()

        self.assertEqual(3, len(extensions))
        self.assertEqual([98, 99, 100], extensions)

        # Make sure we pick up the correct contribution via the plugin.
        extensions = a.x[:]
        extensions.sort()

        self.assertEqual(3, len(extensions))
        self.assertEqual([98, 99, 100], extensions)

        # Make sure we got a trait event telling us that the contributions
        # to the extension point have been changed.
        self.assertEqual(a, listener.obj)
        self.assertEqual("x_items", listener.trait_name)
        self.assertEqual([], listener.new.added)
        self.assertEqual([1, 2, 3], listener.new.removed)
        self.assertEqual(0, listener.new.index.start)
        self.assertEqual(3, listener.new.index.stop)

    def test_assign_empty_list_with_observe(self):
        """ assign an empty list to a plugin triggers a list change event."""

        a = PluginA()
        b = PluginB()
        c = PluginC()

        events = []
        a.observe(events.append, "x:items")

        application = TestApplication(plugins=[a, b, c])
        application.start()

        # Assign an empty list to one of the plugin's contributions.
        b.x = []

        # then
        self.assertEqual(len(events), 1)
        event, = events
        self.assertEqual(event.object, a.x)
        self.assertEqual(event.added, [])
        self.assertEqual(event.removed, [1, 2, 3])
        self.assertEqual(event.index, 0)

    def test_assign_non_empty_list(self):
        """ assign non-empty list """

        a = PluginA()
        a.on_trait_change(listener, "x_items")
        b = PluginB()
        c = PluginC()

        application = TestApplication(plugins=[a, b, c])
        application.start()

        # Keep the old values for later slicing check
        source_values = list(a.x)

        # Assign a non-empty list to one of the plugin's contributions.
        b.x = [2, 4, 6, 8]

        # Make sure we pick up the new contribution via the application.
        extensions = application.get_extensions("a.x")
        extensions.sort()

        self.assertEqual(7, len(extensions))
        self.assertEqual([2, 4, 6, 8, 98, 99, 100], extensions)

        # Make sure we pick up the new contribution via the plugin.
        extensions = a.x[:]
        extensions.sort()

        self.assertEqual(7, len(extensions))
        self.assertEqual([2, 4, 6, 8, 98, 99, 100], extensions)

        # Make sure we got a trait event telling us that the contributions
        # to the extension point have been changed.
        self.assertEqual(a, listener.obj)
        self.assertEqual("x_items", listener.trait_name)
        self.assertEqual([2, 4, 6, 8], listener.new.added)
        self.assertEqual([1, 2, 3], listener.new.removed)

        # The removed entry should match what the old values say
        self.assertEqual(
            listener.new.removed, source_values[listener.new.index]
        )

        # If we use the index and apply the changes to the old list, we should
        # recover the new list
        source_values[listener.new.index] = listener.new.added
        self.assertEqual(source_values, application.get_extensions("a.x"))

        self.assertEqual(0, listener.new.index.start)
        self.assertEqual(3, listener.new.index.stop)

    def test_assign_non_empty_list_with_observe(self):
        """ assign non-empty list """

        a = PluginA()
        b = PluginB()
        c = PluginC()

        events = []
        a.observe(events.append, "x:items")

        application = TestApplication(plugins=[a, b, c])
        application.start()

        # Assign a non-empty list to one of the plugin's contributions.
        b.x = [2, 4, 6, 8]

        # then
        self.assertEqual(len(events), 1)
        event, = events
        self.assertEqual(event.object, a.x)
        self.assertEqual(event.index, 0)
        self.assertEqual(event.added, [2, 4, 6, 8])
        self.assertEqual(event.removed, [1, 2, 3])

    def test_add_plugin(self):
        """ add plugin """

        a = PluginA()
        a.on_trait_change(listener, "x_items")
        b = PluginB()
        c = PluginC()

        # Start off with just two of the plugins.
        application = TestApplication(plugins=[a, b])
        application.start()

        # Make sure we can get the contributions via the application.
        extensions = application.get_extensions("a.x")
        extensions.sort()

        self.assertEqual(3, len(extensions))
        self.assertEqual([1, 2, 3], extensions)

        # Make sure we can get the contributions via the plugin.
        extensions = a.x[:]
        extensions.sort()

        self.assertEqual(3, len(extensions))
        self.assertEqual([1, 2, 3], extensions)

        # Now add the other plugin.
        application.add_plugin(c)

        # Make sure we can get the contributions via the application.
        extensions = application.get_extensions("a.x")
        extensions.sort()

        self.assertEqual(6, len(extensions))
        self.assertEqual([1, 2, 3, 98, 99, 100], extensions)

        # Make sure we can get the contributions via the plugin.
        extensions = a.x[:]
        extensions.sort()

        self.assertEqual(6, len(extensions))
        self.assertEqual([1, 2, 3, 98, 99, 100], extensions)

        # Make sure we got a trait event telling us that the contributions
        # to the extension point have been changed.
        self.assertEqual(a, listener.obj)
        self.assertEqual("x_items", listener.trait_name)
        self.assertEqual([98, 99, 100], listener.new.added)
        self.assertEqual([], listener.new.removed)
        self.assertEqual(3, listener.new.index)

    def test_add_plugin_with_observe(self):
        """ add plugin with observe """

        a = PluginA()
        b = PluginB()
        c = PluginC()

        events = []
        a.observe(events.append, "x:items")

        # Start off with just two of the plugins.
        application = TestApplication(plugins=[a, b])
        application.start()

        # Now add the other plugin.
        application.add_plugin(c)

        # then
        self.assertEqual(len(events), 1)
        event, = events
        self.assertEqual(event.object, a.x)
        self.assertEqual(event.index, 3)
        self.assertEqual(event.added, [98, 99, 100])
        self.assertEqual(event.removed, [])

    def test_remove_plugin(self):
        """ remove plugin """

        a = PluginA()
        a.on_trait_change(listener, "x_items")
        b = PluginB()
        c = PluginC()

        application = TestApplication(plugins=[a, b, c])
        application.start()

        # Make sure we can get the contributions via the application.
        extensions = application.get_extensions("a.x")
        extensions.sort()

        self.assertEqual(6, len(extensions))
        self.assertEqual([1, 2, 3, 98, 99, 100], extensions)

        # Make sure we can get the contributions via the plugin.
        extensions = a.x[:]
        extensions.sort()

        self.assertEqual(6, len(extensions))
        self.assertEqual([1, 2, 3, 98, 99, 100], extensions)

        # Now remove one plugin.
        application.remove_plugin(b)

        # Make sure we can get the contributions via the application.
        extensions = application.get_extensions("a.x")
        extensions.sort()

        self.assertEqual(3, len(extensions))
        self.assertEqual([98, 99, 100], extensions)

        # Make sure we can get the contributions via the plugin.
        extensions = a.x[:]
        extensions.sort()

        self.assertEqual(3, len(extensions))
        self.assertEqual([98, 99, 100], extensions)

        # Make sure we got a trait event telling us that the contributions
        # to the extension point have been changed.
        self.assertEqual(a, listener.obj)
        self.assertEqual("x_items", listener.trait_name)
        self.assertEqual([], listener.new.added)
        self.assertEqual([1, 2, 3], listener.new.removed)
        self.assertEqual(0, listener.new.index)

    def test_remove_plugin_with_observe(self):
        """ remove plugin with observe """

        a = PluginA()
        b = PluginB()
        c = PluginC()

        events = []
        a.observe(events.append, "x:items")

        # Start off with just two of the plugins.
        application = TestApplication(plugins=[a, b, c])
        application.start()

        # Now remove one plugin.
        application.remove_plugin(b)

        # then
        self.assertEqual(len(events), 1)
        event, = events
        self.assertEqual(event.object, a.x)
        self.assertEqual(event.index, 0)
        self.assertEqual(event.added, [])
        self.assertEqual(event.removed, [1, 2, 3])

    def test_race_condition(self):
        """ Test the extension point being modified before the application
        starts, changes before starting the application are not notified.
        """
        a = PluginA()
        b = PluginB()
        c = PluginC()
        application = TestApplication(plugins=[a, b, c])

        events = []
        a.observe(events.append, "x:items")

        # This sets the cache.
        self.assertEqual(a.x, [1, 2, 3, 98, 99, 100])

        # Now we mutate the registry, but the application has not started.
        b.x = [4, 5, 6]

        # then
        # The values are not synchronized.
        self.assertEqual(a.x, [1, 2, 3, 98, 99, 100])

        # application has not started, no events.
        self.assertEqual(len(events), 0)

        # Now we start the application, which connects the listener.
        application.start()

        # then
        self.assertEqual(a.x, [4, 5, 6, 98, 99, 100])

        # Change the value again.
        b.x = [1, 2]

        # then
        self.assertEqual(a.x, [1, 2, 98, 99, 100])

        # The mutation occurred before application starting is not reported.
        self.assertEqual(len(events), 1)
        event, = events
        self.assertEqual(event.object, a.x)
        self.assertEqual(event.index, 0)
        self.assertEqual(event.added, [1, 2])
        self.assertEqual(event.removed, [4, 5, 6])


class TestExtensionPointChangedEvent(unittest.TestCase):
    """ Test ExtensionPointChangedEvent object."""

    def test_extension_point_change_event_str_representation(self):
        """ test string representation of the ExtensionPointChangedEvent class
        """
        desired_repr = ("ExtensionPointChangedEvent(extension_point_id={}, "
                        "index=0, removed=[], added=[])")
        ext_pt_changed_evt = ExtensionPointChangedEvent(extension_point_id=1)
        self.assertEqual(desired_repr.format(1), str(ext_pt_changed_evt))
        self.assertEqual(desired_repr.format(1), repr(ext_pt_changed_evt))



# Enthought library imports.
from pyface.tasks.api import TraitsDockPane
from traits.api import HasTraits, Instance
from traitsui.api import Item, View


class ModelConfigPane(TraitsDockPane):
    """ A simple dock pane for editing an attractor model's configuration
        options.
    """

    #### 'ITaskPane' interface ################################################

    id = "example.attractors.model_config_pane"
    name = "Model Configuration"

    #### 'ModelConfigPane' interface ##########################################

    model = Instance(HasTraits)

    view = View(
        Item("pane.model", style="custom", show_label=False), resizable=True
    )
